@@ -2,6 +2,7 @@ from __future__ import annotations
 
 import dataclasses
 import functools
+import json
 import os
 import re
 import secrets
@@ -18,7 +19,7 @@ from pip._vendor.pkg_resources import Requirement as PackageRequirement
 from pip._vendor.pkg_resources import RequirementParseError, safe_name
 
 from pdm._types import RequirementDict
-from pdm.exceptions import ExtrasError, RequirementError
+from pdm.exceptions import ExtrasWarning, RequirementError
 from pdm.models.markers import Marker, get_marker, split_marker_extras
 from pdm.models.pip_shims import (
     InstallRequirement,
@@ -79,6 +80,7 @@ class Requirement:
     extras: Sequence[str] | None = None
     specifier: SpecifierSet | None = None
     editable: bool = False
+    prerelease: bool = False
 
     def __post_init__(self) -> None:
         self.requires_python = (
@@ -157,6 +159,30 @@ class Requirement:
             kwargs["specifier"] = get_specifier(version)
         return cls(**kwargs)
 
+    @classmethod
+    def from_dist(cls, dist: Distribution) -> "Requirement":
+        direct_url_json = dist.read_text("direct_url.json")
+        if direct_url_json is not None:
+            direct_url = json.loads(direct_url_json)
+            data = {
+                "name": dist.metadata["Name"],
+                "url": direct_url.get("url"),
+                "editable": direct_url.get("dir_info", {}).get("editable"),
+                "subdirectory": direct_url.get("subdirectory"),
+            }
+            if "vcs_info" in direct_url:
+                vcs_info = direct_url["vcs_info"]
+                data.update(
+                    url=f"{vcs_info['vcs']}+{direct_url['url']}",
+                    ref=vcs_info.get("requested_revision"),
+                    revision=vcs_info.get("commit_id"),
+                )
+                return VcsRequirement.create(**data)
+            return FileRequirement.create(**data)
+        return NamedRequirement.create(
+            name=dist.metadata["Name"], version=f"=={dist.version}"
+        )
+
     @classmethod
     def from_req_dict(cls, name: str, req_dict: RequirementDict) -> "Requirement":
         if isinstance(req_dict, str):  # Version specifier only.
@@ -199,11 +225,14 @@ class Requirement:
 
     def as_ireq(self, **kwargs: Any) -> InstallRequirement:
         line_for_req = self.as_line()
-        if self.editable:
-            line_for_req = line_for_req[3:].strip()
-            ireq = install_req_from_editable(line_for_req, **kwargs)
-        else:
-            ireq = install_req_from_line(line_for_req, **kwargs)
+        try:
+            if self.editable:
+                line_for_req = line_for_req[3:].strip()
+                ireq = install_req_from_editable(line_for_req, **kwargs)
+            else:
+                ireq = install_req_from_line(line_for_req, **kwargs)
+        except Exception as e:
+            raise RequirementError(e)
         ireq.req = self  # type: ignore
         return ireq
 
@@ -216,11 +245,11 @@ class Requirement:
             "marker": get_marker(req.marker),
         }
         if getattr(req, "url", None):
-            link = Link(req.url)
+            link = Link(cast(str, req.url))
             klass = VcsRequirement if link.is_vcs else FileRequirement
             return klass(url=req.url, **kwargs)  # type: ignore
         else:
-            return NamedRequirement(**kwargs)
+            return NamedRequirement(**kwargs)  # type: ignore
 
     def _format_marker(self) -> str:
         if self.marker:
@@ -280,9 +309,10 @@ class FileRequirement(Requirement):
     def _parse_url(self) -> None:
         if not self.url:
             if self.path:
-                self.url = path_to_url(
-                    self.path.as_posix().replace("${PROJECT_ROOT}", ".")
-                )
+                self.url = path_to_url(self.path.resolve().as_posix())
+                if not self.path.is_absolute():
+                    project_root = Path(".").resolve().as_posix().lstrip("/")
+                    self.url = self.url.replace(project_root, "${PROJECT_ROOT}")
         else:
             try:
                 self.path = Path(
@@ -411,25 +441,29 @@ class VcsRequirement(FileRequirement):
 
 
 def filter_requirements_with_extras(
-    requirement_lines: list[str],
-    extras: Sequence[str],
+    project_name: str, requirement_lines: list[str], extras: Sequence[str]
 ) -> list[str]:
+    """Filter the requirements with extras.
+    If extras are given, return those with matching extra markers.
+    Otherwise, return those without extra markers.
+    """
     result: list[str] = []
-    extras_in_meta: list[str] = []
+    extras_in_meta: set[str] = set()
     for req in requirement_lines:
         _r = parse_requirement(req)
-        if not _r.marker:
-            result.append(req)
+        if _r.marker:
+            elements, rest = split_marker_extras(str(_r.marker))
+            if elements:
+                extras_in_meta.update(elements)
+                _r.marker = Marker(rest) if rest else None
         else:
-            elements, rest = split_marker_extras(_r.marker)
-            extras_in_meta.extend(elements)
-            _r.marker = Marker(str(rest)) if rest else None
-            if not elements or set(extras) & set(elements):
-                result.append(_r.as_line())
+            elements = set()
+        if extras and not elements.isdisjoint(extras) or not (extras or elements):
+            result.append(_r.as_line())
 
     extras_not_found = [e for e in extras if e not in extras_in_meta]
     if extras_not_found:
-        warnings.warn(ExtrasError(extras_not_found))
+        warnings.warn(ExtrasWarning(project_name, extras_not_found))
 
     return result
 
